/**
 * @file
 */

/*
Copyright (C) 2002-2013 UFO: Alien Invasion.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "../../../../client.h"
#include "../../../../cgame/cl_game.h"
#include "e_event_endroundannounce.h"

/**
 * @brief Announces that a player ends his turn
 * @param[in] self Pointer to the event structure that is currently executed
 * @param[in] msg The message buffer to read from
 * @sa CL_DoEndRound
 * @note event EV_ENDROUNDANNOUNCE
 * @todo Build into hud
 */
void CL_EndRoundAnnounce (const eventRegister_t* self, dbuffer* msg)
{
	/* get the needed values */
	const int playerNum = NET_ReadByte(msg);
	const int team = NET_ReadByte(msg);

	GAME_EndRoundAnnounce(playerNum, team);
}
