/**
 * @file
 */

/*
 Copyright (C) 2002-2013 UFO: Alien Invasion.
 All Rights Reserved.

 This file is part of UFORadiant.

 UFORadiant is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 UFORadiant is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GtkRadiant; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef RADIANT_SHARED_H
#define RADIANT_SHARED_H

#ifdef HAVE_CONFIG_H
#include "../../../../config.h"
#endif

#include "../../../shared/defines.h"

#endif
