#pragma once

#define HARD_LINKED_GAME 1
#define HAVE_THEORA_THEORA_H 1
#define HAVE_SYS_UTSNAME_H 1
#define HAVE_MXML_H 1
#define PKGDATADIR "."
#define PKGLIBDIR "."
/*
#define PARANOID
#define HARD_LINKED_CGAME
#define USE_SIGNALS 1
#define LOCALEDIR "."
#define HAVE_XVID_H 1
#define HAVE_EXECINFO_H 1
#define HAVE_BFD_H 1
#define HAVE_SYS_UTSNAME_H 1
#define HAVE_LINK_H 1
#define HAVE_CUNIT_BASIC_H 1
*/
