#pragma once

namespace ui
{
	namespace icons
	{
		namespace
		{
			/* CONSTANTS */
			static const std::string ICON = "icon.png";

			static const std::string ICON_OPEN = "file_open.png";
			static const std::string ICON_SAVE = "file_save.png";
			static const std::string ICON_UNDO = "undo.png";
			static const std::string ICON_REDO = "redo.png";

			static const std::string ICON_LEVEL1 = "ufoai_level1.png";
			static const std::string ICON_LEVEL2 = "ufoai_level2.png";
			static const std::string ICON_LEVEL3 = "ufoai_level3.png";
			static const std::string ICON_LEVEL4 = "ufoai_level4.png";
			static const std::string ICON_LEVEL5 = "ufoai_level5.png";
			static const std::string ICON_LEVEL6 = "ufoai_level6.png";
			static const std::string ICON_LEVEL7 = "ufoai_level7.png";
			static const std::string ICON_LEVEL8 = "ufoai_level8.png";

			static const std::string ICON_HELP = "helpicon.png";

			static const std::string ICON_GRID_TOGGLE = "grid_toggle.png";
			static const std::string ICON_GRID_UP = "grid_up.png";
			static const std::string ICON_GRID_DOWN = "grid_down.png";
			static const std::string ICON_GRID_SNAP = "grid_snap.png";

			static const std::string ICON_BRUSH_FLIPX = "brush_flipx.png";
			static const std::string ICON_BRUSH_ROTATEX = "brush_rotatex.png";
			static const std::string ICON_BRUSH_FLIPY = "brush_flipy.png";
			static const std::string ICON_BRUSH_ROTATEY = "brush_rotatey.png";
			static const std::string ICON_BRUSH_FLIPZ = "brush_flipz.png";
			static const std::string ICON_BRUSH_ROTATEZ = "brush_rotatez.png";

			static const std::string ICON_SELECT_TOUCHING = "selection_selecttouching.png";
			static const std::string ICON_SELECT_INSIDE = "selection_selectinside.png";
			static const std::string ICON_SELECT_ENTITIES = "selection_selectentities.png";
			static const std::string ICON_SELECT_COMPLETE = "selection_selectcompletetall.png";
			static const std::string ICON_SELECT_SAME_TEXTURE = "selection_selectallsametex.png";

			static const std::string ICON_CSG_SUBTRACT = "selection_csgsubtract.png";
			static const std::string ICON_CSG_MERGE = "selection_csgmerge.png";
			static const std::string ICON_CSG_MAKE_HOLLOW = "selection_makehollow.png";

			static const std::string ICON_MODIFY_VERTICES = "modify_vertices.png";
			static const std::string ICON_MODIFY_EDGES = "modify_edges.png";
			static const std::string ICON_MODIFY_FACES = "modify_faces.png";

			static const std::string ICON_VIEW_CLIPPER = "view_clipper.png";

			static const std::string ICON_FILTER_NODRAW = "filter_nodraw.png";
			static const std::string ICON_FILTER_ACTORCLIP = "filter_actorclip.png";
			static const std::string ICON_FILTER_WEAPONCLIP = "filter_weaponclip.png";

			static const std::string ICON_VIEW_CHANGE = "view_change.png";

			static const std::string ICON_SELECT_MOUSE_TRANSLATE = "select_mousetranslate.png";
			static const std::string ICON_SELECT_MOUSE_ROTATE = "select_mouserotate.png";
			static const std::string ICON_SELECT_MOUSE_SCALE = "select_mousescale.png";
			static const std::string ICON_SELECT_MOUSE_RESIZE = "select_mouseresize.png";

			static const std::string ICON_REFRESH_MODELS = "refresh_models.png";
			static const std::string ICON_BACKGROUND = "background.png";
			static const std::string ICON_MATERIALS_GENERATE = "material_generate.png";
			static const std::string ICON_PATHFINDING = "pathfinding.png";
			static const std::string ICON_TEXTURE_LOCK = "texture_lock.png";

			static const std::string ICON_VIEW_CLIPPING = "view_cubicclipping.png";

			static const std::string ICON_DRAW_BBOX = "iconDrawBBox.png";
			static const std::string ICON_MD2 = "model16red.png";
			static const std::string ICON_MD3 = "model16green.png";
			static const std::string ICON_OBJ = "model16blue.png";
			static const std::string ICON_SKIN = "skin16.png";
			static const std::string ICON_FOLDER = "folder16.png";
			static const std::string ICON_SOUND = "icon_sound.png";
			static const std::string ICON_ADD_SOUND = "icon_sound.png";

			static const std::string ICON_UP = "arrow_up.png";
			static const std::string ICON_DOWN = "arrow_down.png";

			static const std::string ICON_COLOR = "editcolour16.png";

			static const std::string ICON_ADD_LIGHT = "cmenu_add_light.png";
			static const std::string ICON_ADD_MODEL = "cmenu_add_model.png";
			static const std::string ICON_ADD_ENTITY = "cmenu_add_entity.png";
			static const std::string ICON_CONNECT_ENTITIES = "cmenu_connect_entities.png";
			static const std::string ICON_FIT_TEXTURE = "cmenu_fit_texture.png";
			static const std::string ICON_GENERATE_MATERIALS = "cmenu_generate_materials.png";
			static const std::string ICON_GENERATE_TERRAIN = "cmenu_generate_terrain.png";

			static const std::string ICON_FIND_AND_REPLACE = "texwindow_findandreplace.png";

			static const std::string ICON_LOGO = "logo.png";

			static const std::string ICON_WINDOW_REGULAR = "window_regular.png";
			static const std::string ICON_WINDOW_SPLIT = "window_split.png";

			static const std::string ICON_PARTICLE_MAP = "particle_map.png";
			static const std::string ICON_PARTICLE_WEAPON = "particle_weapon.png";

			static const std::string ICON_MODEL_SHOWNORMALS = "model_normals.png";
		}
	}
}
