/**
 * @file
 */

/*
Copyright (C) 1999-2007 id Software, Inc. and contributors.
For a list of contributors, see the accompanying CONTRIBUTORS file.

This file is part of GtkRadiant.

GtkRadiant is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

GtkRadiant is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GtkRadiant; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#pragma once

#include "../../../shared/mathlib.h"

typedef struct {
	vec3_t verts[3];
	vec3_t normals[3];
	vec2_t texcoords[3];
} triangle_t;

typedef struct {
	char name[100];
	char materialname[100];
	triangle_t* triangles;
	int numtriangles;
} polyset_t;

void ASE_Load(const char* filename, bool verbose);
polyset_t* ASE_GetSurfaceAnimation(int whichSurface);
int ASE_GetNumSurfaces(void);
const char* ASE_GetSurfaceName(int ndx);
void ASE_Free(void);
