/**
 * @file
 */

/*
Copyright (C) 2002-2013 UFO: Alien Invasion.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#pragma once

struct uiNode_t;

typedef struct uiFont_s {
	char* name;
	int size;
	char* style;
	char* path;
} uiFont_t;

/* will return the size and the path for each font */
const char* UI_GetFontFromNode(uiNode_t const* node);
const uiFont_t* UI_GetFontByID(const char* fontID);
/* this is the function where all the sdl_ttf fonts are parsed */
bool UI_ParseFont(const char* name, const char** text);
void UI_InitFonts(void);
void UI_FontShutdown(void);
int UI_FontGetHeight(const char* font);
