/**
 * @mainpage UFO:Alien Invasion
 *
 * The latest version of UFO:Alien Invasion can be found at:
 *     http://sourceforge.net/projects/ufoai/
 *
 * UFO:AI; A strategy game featuring tactical combat against hostile alien
 * forces which are about to infiltrate earth at this very moment.
 *
 * http://cia.vc/stats/project/ufoai
 * http://ufoai.sourceforge.net/
 * http://colabti.org/irclogger/irclogger_logs/ufoai
 * @author UFO:AI Team
 */

#pragma once
